/*
 * Licensed Materials - Property of IBM
 *
 * com.ibm.etools.c
 *
 *  Copyright IBM Corporation 2004. All Rights Reserved. 
 *
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.etools.c.importer;

import com.ibm.etools.c.CField;
import com.ibm.etools.c.CStruct;
import com.ibm.etools.c.datatypes.CBitField;

/**
 * CBitFieldHelper is a helper class for CTypeDescriptorHelper. It is used to
 * calculate the storage layout for structures with bitfields.
 * 
 * The bitfield layout rule is totally implementation dependent. Different
 * compilers do it in a different way. This class provides full support for two
 * compilers: MSVC and GCC. The GCC rule is more efficient and results in a more
 * compact storage layout. Many compilers on Unix platform use a rule similar to
 * the GCC rule. So we use the GCC rule as the default for an unsupported
 * compiler.
 */
public class CBitFieldHelper {

    public static final String copyright = "Licensed Material - Property of IBM\n com.ibm.etools.c\n(C) Copyright IBM Corp. 2004 - All Rights Reserved.\nUS Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp."; //$NON-NLS-1$

    private static CBitFieldHelper _instance = new CBitFieldHelper();

    public static int RULE_MSVC = 0;

    public static int RULE_GCC = 1;

    private static int _rule;

    private static CStruct _currentStruct;

    private static CStruct _lastStruct;

    private static CBitField _currentBitField;

    private static int _lastBitFieldIndex = -1;

    private static int _currentBitFieldIndex;

    private static int _currentStructSize;

    private static int _numberOfRemainingBitsInLastByte = 0;

    private static boolean _isUnnamedBitField;

    private static boolean _isZeroSizeBitField;

    private static int _baseSizeOfLastField = -1;

    private static int _baseSizeOfThisBitField;

    private static int _numberOfRemainingBits;

    private static int _sizeOfCurrentBitfield;

    private static int _offsetOfCurrentBitField;

    private static int _alignment;

    /**
     * Constructor
     */
    private CBitFieldHelper() {
    }

    /**
     * Return the only instance of CBitFieldHelper
     */
    public static CBitFieldHelper instance() {
        return _instance;
    }

    /**
     * Set the bitfield layout rule
     */
    public static void setRule() {
        String compilerName = CPreferenceStore.getCompilerName();
        if (compilerName != null && compilerName.toLowerCase().equals("msvc"))
            _rule = RULE_MSVC;
        else
            _rule = RULE_GCC;
    }

    /**
     * Parse the given bitfield in a given structure.
     * 
     * @param astruct
     *            The structure which contains the bitfield
     * @param afield
     *            The field whose type is a bitfield
     * @param index
     *            The index of the bitfield within the structure
     * @param size
     *            The current size of the structure before calculating the
     *            bitfield
     */
    public static void parse(CStruct astruct, CField afield, int index, int size) throws CException {
        _currentStruct = astruct;
        _currentBitField = (CBitField) CQuery.typeOf(afield);
        _currentBitFieldIndex = index;
        _currentStructSize = size;

        _baseSizeOfThisBitField = CTypeDescriptorHelper.sizeof(_currentBitField.getBaseType());
        _sizeOfCurrentBitfield = _currentBitField.getSize().intValue();

        if (_sizeOfCurrentBitfield == 0)
            _isZeroSizeBitField = true;
        else
            _isZeroSizeBitField = false;

        if (_currentStruct != _lastStruct) {
            _numberOfRemainingBits = 0;
            _numberOfRemainingBitsInLastByte = 0;
        }

        String bitfieldName = afield.getName();
        if (bitfieldName == null || bitfieldName.equals(""))
            _isUnnamedBitField = true;
        else
            _isUnnamedBitField = false;

        if (_rule == RULE_MSVC)
            parseIt_MSVC();
        else
            parseIt_GCC();

        _lastStruct = _currentStruct;
        _lastBitFieldIndex = _currentBitFieldIndex;
    }

    /**
     * Calculate storage layout for the current bitfield using the MSVC rule
     * 
     * Bitfield layout mechanism for MSVC: 1. Adjacent bitfields with the same
     * base type are packed together 2. Adjacent bitfields with different base
     * types are not packed 3. The base type of the bitfield affects its
     * alignment 4. Unnamed bitfields affect alignment 5. Zero size bitfield
     * following a non-bitfield has no effect on storage requirement 6. Zero
     * size bitfield following a bitfield indicates that the next bitfield must
     * start at a new storage unit
     */
    private static void parseIt_MSVC() throws CException {
        boolean lastFieldIsBitField = false;
        if (_currentStruct == _lastStruct && _lastBitFieldIndex >= 0 && _currentBitFieldIndex == _lastBitFieldIndex + 1)
            lastFieldIsBitField = true;
        else
            _numberOfRemainingBits = 0;

        _alignment = CTypeDescriptorHelper.alignmentOf(_currentBitField);
        int radius = _currentStructSize % _alignment;

        int paddingBytes = 0;
        int allocBytes = 0;

        if (_isZeroSizeBitField) {
            if (lastFieldIsBitField) {
                if (radius == 0)
                    paddingBytes = 0;
                else
                    paddingBytes = _baseSizeOfThisBitField - radius;
            } else {
                _alignment = 1;
            }

            _offsetOfCurrentBitField = _currentStructSize + paddingBytes;
            _numberOfRemainingBits = 0;
        } else if (lastFieldIsBitField && _baseSizeOfThisBitField == _baseSizeOfLastField) {
            int bitsOfThisBitField = _currentBitField.getSize().intValue();
            if (bitsOfThisBitField <= _numberOfRemainingBits) {
                _offsetOfCurrentBitField = _currentStructSize * 8 - _numberOfRemainingBits;
                allocBytes = 0;
                _numberOfRemainingBits -= bitsOfThisBitField;
            } else {
                paddingBytes = 0;
                _offsetOfCurrentBitField = _currentStructSize * 8;
                allocBytes = _baseSizeOfThisBitField;
                _numberOfRemainingBits = allocBytes * 8 - bitsOfThisBitField;
            }
        } else {
            radius = _currentStructSize % _alignment;
            if (radius == 0) {
                _offsetOfCurrentBitField = 8 * _currentStructSize;
                paddingBytes = 0;
            } else {
                _offsetOfCurrentBitField = 8 * (_currentStructSize + _alignment - radius);
                paddingBytes = _alignment - radius;
            }

            allocBytes = _baseSizeOfThisBitField;
            _numberOfRemainingBits = allocBytes * 8 - _currentBitField.getSize().intValue();
            _baseSizeOfLastField = _baseSizeOfThisBitField;

        }

        _currentStructSize += paddingBytes + allocBytes;
    }

    /**
     * Parse the bitfield using the layout rule of gcc
     * 
     * Bitfield layout mechanism for gcc: 1. Adjacent bitfields with different
     * base types are packed together 2. The base type of the bitfield affects
     * the struct alignment, but does not affect the size and alignment of this
     * bitfield 3. Unnamed bitfield does not affect the struct alignment 4. Zero
     * size bitfield indicates that the next struct field must start at an
     * address that is a multiple of the bitfield base type.
     */
    private static void parseIt_GCC() throws CException {

        boolean lastFieldIsBitField = false;
        if (_currentStruct == _lastStruct && _lastBitFieldIndex >= 0 && _currentBitFieldIndex == (_lastBitFieldIndex + 1)) {
            lastFieldIsBitField = true;
        } else {
            lastFieldIsBitField = false;
            _numberOfRemainingBitsInLastByte = 0;
        }

        if (_isUnnamedBitField)
            _alignment = 1;
        else
            _alignment = CTypeDescriptorHelper.alignmentOf(_currentBitField);

        int radius = _currentStructSize % _baseSizeOfThisBitField;

        int baseBits = 0;
        if (radius == 0)
            baseBits = _numberOfRemainingBitsInLastByte;
        else
            baseBits = _numberOfRemainingBitsInLastByte + (_baseSizeOfThisBitField - radius) * 8;

        int paddingBytes = 0;
        int allocBytes = 0;

        if (_isZeroSizeBitField) {
            if (radius == 0)
                paddingBytes = 0;
            else
                paddingBytes = _baseSizeOfThisBitField - radius;

            lastFieldIsBitField = false;
            _numberOfRemainingBitsInLastByte = 0;
        } else if (_sizeOfCurrentBitfield <= _numberOfRemainingBitsInLastByte) {
            _offsetOfCurrentBitField = _currentStructSize * 8 - _numberOfRemainingBitsInLastByte;
            _numberOfRemainingBitsInLastByte -= _sizeOfCurrentBitfield;
        } else if (_sizeOfCurrentBitfield < baseBits) {
            _offsetOfCurrentBitField = _currentStructSize * 8 - _numberOfRemainingBitsInLastByte;
            int extraBits = _sizeOfCurrentBitfield - _numberOfRemainingBitsInLastByte;
            if (extraBits % 8 == 0) {
                allocBytes = extraBits / 8;
                ;
                _numberOfRemainingBitsInLastByte = 0;
            } else {
                allocBytes = extraBits / 8 + 1;
                _numberOfRemainingBitsInLastByte = 8 - extraBits % 8;
            }
        } else {
            if (radius == 0)
                paddingBytes = 0;
            else
                paddingBytes = _baseSizeOfThisBitField - radius;

            _offsetOfCurrentBitField = 8 * (_currentStructSize + paddingBytes);
            if (_sizeOfCurrentBitfield % 8 == 0) {
                allocBytes = _sizeOfCurrentBitfield / 8;
                _numberOfRemainingBitsInLastByte = 0;
            } else {
                allocBytes = _sizeOfCurrentBitfield / 8 + 1;
                _numberOfRemainingBitsInLastByte = 8 - _sizeOfCurrentBitfield % 8;
            }
        }

        _currentStructSize += paddingBytes + allocBytes;

    }

    /**
     * Return the size in bits of the current bitfield
     */
    public static int getSize() {
        return _sizeOfCurrentBitfield;
    }

    /**
     * Return the alignment of the current bitfield
     */
    public static int getAlignment() {
        return _alignment;
    }

    /**
     * Return the current structure size after taking the current bitfield into
     * account
     */
    public static int getCurrentStructSize() {
        return _currentStructSize;
    }

    /**
     * Return the offset in bits of the current bitfield
     */
    public static int getOffset() {
        return _offsetOfCurrentBitField;
    }

}