/*
 * Licensed Materials - Property of IBM
 *
 * com.ibm.etools.c
 *
 *  Copyright IBM Corporation 2004. All Rights Reserved. 
 *
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.etools.c.importer;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;

import org.eclipse.cdt.core.parser.CodeReader;
import org.eclipse.cdt.core.parser.IParser;
import org.eclipse.cdt.core.parser.IProblem;
import org.eclipse.cdt.core.parser.IScanner;
import org.eclipse.cdt.core.parser.IScannerInfo;
import org.eclipse.cdt.core.parser.ParserFactory;
import org.eclipse.cdt.core.parser.ParserFactoryError;
import org.eclipse.cdt.core.parser.ParserLanguage;
import org.eclipse.cdt.core.parser.ParserMode;
import org.eclipse.cdt.core.parser.ScannerInfo;
import org.eclipse.cdt.core.parser.ast.IASTCompilationUnit;
import org.eclipse.cdt.cpp.miners.parser.dstore.ParserSchema;
import org.eclipse.cdt.dstore.core.client.ClientConnection;
import org.eclipse.cdt.dstore.core.model.DE;
import org.eclipse.cdt.dstore.core.model.DataElement;
import org.eclipse.cdt.dstore.core.model.DataStore;
import org.eclipse.cdt.dstore.core.model.DataStoreAttributes;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceImpl;

import com.ibm.etools.c.CStructured;
import com.ibm.etools.c.CTypedef;
import com.ibm.etools.c.impl.CPackageImpl;
import com.ibm.etools.c.importer.AST.CParseCallback;
import com.ibm.etools.c.messages.Messages;
import com.ibm.etools.c.plugin.CPlugin;

/**
 * @author seto
 * @author Bernhard Jakob, IBM
 */
public class CImporterResource extends XMIResourceImpl {

    public static final String copyright = "Licensed Material - Property of IBM\n com.ibm.etools.c\n(C) Copyright IBM Corp. 2004 - All Rights Reserved.\nUS Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp."; //$NON-NLS-1$

    private String _uri;

    private ArrayList _includePaths;

    private CModelConverter _currentModelConverter;

    private DataStore _dataStore = null;

    // Factories for the C TDLang model
    //private CFactory _factory;
    //private DatatypesFactory _datatypesFactory;

    /**
     *  
     */
    public CImporterResource() {
        this(null);
    }

    /**
     * @param arg0
     */
    public CImporterResource(URI arg0) {
        super(arg0);
        // Initialize the C Package if running from the command line
        // If running within the Eclipse workbench, the packages are initialized
        // via plugin.xml.
        //
        if (!Platform.isRunning()) {
            CPackageImpl.init();
        }

        _currentModelConverter = CModelConverter.createInstance();
        _includePaths = new ArrayList();
    }

    // /**
    //  * Return the single instance of the CImporter
    //  */
    // public static CImporter instance() {
    //
    //	if(_thisImporter == null) {
    //		_thisImporter = new CImporter();
    //	}
    //
    //	return _thisImporter;
    // }

    // /**
    //  * Return the C Factory used to create C model elements
    //  */
    // public CFactory getFactory() {
    //   return _factory;
    // }
    //
    // /**
    //  * Return the C DataTypes Factory used to create C datatype elements
    //  */
    // public DatatypesFactory getDatatypesFactory() {
    //   return _datatypesFactory;
    // }

    /**
     * Return the current CModelConverter object
     */
    public CModelConverter getCurrentModelConverter() {
        return _currentModelConverter;
    }

    /**
     * Parse the include paths from a string separated by the pathSeparator and
     * store each path element into an array list.
     * 
     * @param pathString
     *            The string that represents the include paths
     * @param includePaths
     *            The array which stores the parsed result
     */
    private void parseIncludePath(String pathString, ArrayList includePaths) {

        if (pathString == null || pathString.length() == 0)
            return;

        if (pathString.indexOf(File.pathSeparatorChar) == -1)
            includePaths.add(pathString);
        else {
            StringTokenizer tokenizer = new StringTokenizer(pathString, File.pathSeparator);
            while (tokenizer.hasMoreTokens()) {
                includePaths.add(tokenizer.nextToken());
            }
        }

    }

    /**
     * Return an array list that contains all the include files of the given
     * source file
     */
    public ArrayList getIncludeFiles(String fileName) throws Exception {

        String fullName = new String(fileName);
        File file = new File(fileName);
        if (file.exists()) {
            fullName = file.getAbsolutePath();
        } else {
            throw new CException(CResource.getString("_ERROR_File_Not_Exist", fileName));
        }

        CSourceUnit sourceUnit = _currentModelConverter.getSourceUnitFromName(fullName);
        if (sourceUnit != null) {
            return sourceUnit.getIncludeFiles();
        } else {
            return null;
        }
    }

    /**
     * Translate the parsed output saved in the DataElement tree into MOF
     * resource format. The return result is a MOF resource object whose Extent
     * is populated with the parsed object list.
     * 
     * @param parsed_source
     *            com.ibm.cpp.core.model.DataElement a data element representing
     *            the root node of the parsed tree
     * @return a resource object with an extent which is a MOF object list
     */
    protected void doLoad(DataElement parsed_source) throws Exception {

        ArrayList objectList = _currentModelConverter.importSource(parsed_source);

        //System.out.println("Size before expand: " + objectList.size());
        // Call the utility class to generate the full object list
        // The default implementaion is to generate a full object list. The full
        // list can be saved persistently to a disk file. The partial list will
        // suffice if you only use it in memory.
        //
        ArrayList finalObjectList = objectList;
        if (CPreferenceStore.isGenerateFullObjectList()) {
            finalObjectList = CUtility.generateFullObjectList(objectList);
        }

        //System.out.println("Size after expand: " + finalObjectList.size());
        //Extent extent = new com.ibm.etools.emf.ref.impl.ExtentImpl();
        //extent.addAll(finalObjectList);
        //Resource outputResource = new
        // com.ibm.xmi.base.impl.XMIResourceImpl(_uri, extent);
        getContents().addAll(finalObjectList);

    }

    /**
     * Retrieve the named type from the parsed file Only the objects related to
     * that type are added to the returned list
     */
    protected void doLoad(DataElement parsed_source, String typeName) throws Exception {

        if (_currentModelConverter.getSourceUnitFromName(parsed_source.getName()) == null) {
            _currentModelConverter.importSource(parsed_source);
        }

        ArrayList objectList = new ArrayList();
        CTypeTable typeTable = _currentModelConverter.getTypeTable();
        Object modelType = typeTable.getType(typeName);

        if (modelType != null) {
            objectList.add(modelType);
        } else {
            // If we cannot find a type for the given name, try to find it as a
            // function
            Object modelFunction = typeTable.getFunction(typeName);
            if (modelFunction != null) {
                objectList.add(modelFunction);
            } else
                throw new CException(CResource.getString("_ERROR_No_Type_Or_Function", typeName));
        }

        ArrayList finalObjectList = objectList;
        if (CPreferenceStore.isGenerateFullObjectList()) {
            finalObjectList = CUtility.generateFullObjectList(objectList);
        }

        //Extent extent = new com.ibm.etools.emf.ref.impl.ExtentImpl();
        //extent.addAll(finalObjectList);
        //Resource outputResource = new
        // com.ibm.xmi.base.impl.XMIResourceImpl(_uri, extent);

        getContents().addAll(finalObjectList);

    }

    // /**
    //  * This is the only interface in ResourceFactory that we need to implement
    // in order to be
    //  * conformance with the importer framework. It is the entry pointer of the
    // importer.
    //  * This method is indirectly called via a call to
    // ResourceFactory.load(String uri).
    //  *
    //  * Here we initialize the DataStore, invoke the parser on the source file,
    //  * then call the other load(DataElement) method to translate the parsed
    // information into
    //  * MOF format.
    //  *
    //  * @return the Resource object
    //  * @param resources The ResourceSet
    //  * @param uri The URI representing the resource to parse
    //  * @param is The input stream
    //  * @param options The options
    //  */
    // public Resource load(ResourceSet resources, String uri, InputStream is,
    // Object options) throws Exception
    // {
    //
    //   initialize(options);
    //   
    //   _uri = uri;
    //
    //   // Throws an exception if the uri is null.
    //   if (uri == null || uri.length() == 0) {
    //	throw new CException(CResource.getString("_ERROR_URI_Is_Null"));
    //   }
    //
    //   // Get the URIFactory
    //   URIFactory uriFactory = URIFactoryRegister.getFactory();
    //   if (uriFactory == null) {
    //	throw new CException(CResource.getString("_ERROR_URIFactory_Is_Null"));
    //   }
    //
    //   String fileName = null;
    //   String typeName = null;
    //   try {
    //	URI myURI = uriFactory.makeURI(uri);
    //	URL myURL = myURI.getURL();
    //	// If the URL protocol is "platform", we need to resolve it to
    //	// the physical path.
    //	if (myURL.getProtocol().equals("platform")) {
    //	 myURL = Platform.resolve(myURL);
    //	}
    //	fileName = myURL.getFile();
    //	typeName = myURL.getRef();
    //   }
    //   catch (Exception e) {
    //	// If uri is not in a valid format of URI, treat it as a
    //	// regular file name.
    //	fileName = new String(_uri);
    //	int index = _uri.indexOf('#');
    //	if (index > 0) {
    //	 fileName = _uri.substring(0, index);
    //	 typeName = _uri.substring(index+1);
    //	}
    //   }
    //
    //   // Get the absolute path name
    //   File file = new File(fileName);
    //   if (file.exists()) {
    //	 fileName = file.getAbsolutePath();
    //   }
    //   else {
    //	 throw new CException(CResource.getString("_ERROR_File_Not_Exist",
    // fileName));
    //   }
    //
    //   // Create a client connection which will create a DataStore implicitly.
    //   ClientConnection client = new ClientConnection("Import");
    //   _dataStore = client.getDataStore();
    //
    //   // Set the plugin path which is used to locate the data files (e.g.
    // minerFile.dat)
    //   // needed by the cpp model.
    //   // This file is located at %eclipsedir%/plugins/com.ibm.etools.c/.
    //   //
    //   String pluginPath = CPlugin.getInstallationDirectory();
    // //org.eclipse.core.boot.BootLoader.getInstallURL().getFile();
    //// if (baseDir != null) {
    //// pluginPath = baseDir + "plugins/";
    //// }
    //// else {
    //// String eclipseDir = System.getProperty("eclipsedir");
    //// if (eclipseDir == null) {
    //// eclipseDir = "c:/eclipse";
    //// }
    //// pluginPath = eclipseDir + "/plugins/";
    //// }
    //
    //   _dataStore.setAttribute(DataStoreAttributes.A_PLUGIN_PATH, pluginPath);
    //   // the pluginpath now has everything it needs - set the location to blank
    //   _dataStore.setMinersLocation(""); //"com.ibm.etools.c");
    //   client.localConnect();
    //
    //   // Initialize the schema and miners
    //   _dataStore.showTicket(null);
    //   _dataStore.getSchema();
    //   _dataStore.initMiners();
    //
    //   // Create a DataElement to represent the object file we are going to
    // parse.
    //   DataElement parseFile = _dataStore.createObject(_dataStore.getHostRoot(),
    // "file", fileName, fileName);
    //
    //   // Find the miner root under which the parsed information is located.
    //   DataElement minerData =
    // _dataStore.findMinerInformation("com.ibm.cpp.miners.parser.ParseMiner");
    //
    //   // Create a fake project to hold the file
    //   DataElement theProject = _dataStore.createObject(minerData,
    // ParserSchema.Project, "fake project");
    //   DataElement parsedFiles =
    // _dataStore.createObject(theProject,ParserSchema.ParsedFiles,
    // ParserSchema.ParsedFiles);
    //   _dataStore.createObject(theProject,ParserSchema.SystemObjects,
    // ParserSchema.SystemObjects);
    //   _dataStore.createObject(theProject,ParserSchema.SystemObjects,
    // ParserSchema.ProjectObjects);
    //
    //   DataElement projectMinerProject = _dataStore.createObject(minerData,
    // ParserSchema.Project, "fake project2");
    //   _dataStore.createReference(projectMinerProject, theProject,
    // ParserSchema.ParseReference);
    //
    //
    //   // Set the include path
    //   DataElement preferences = _dataStore.createObject(theProject,
    // "preferences", "Preferences");
    //   DataElement includeElement = _dataStore.createObject(preferences,
    // "preferences", "Include Path");
    //   for (int i = 0; i < _includePaths.size(); i++)
    //	_dataStore.createObject(includeElement, "directory",
    // (String)_includePaths.get(i), (String)_includePaths.get(i));
    //
    //   // Invoke the C/C++ parser on the project
    //   ArrayList args = new ArrayList();
    //   args.add(projectMinerProject);
    //   DataElement status = _dataStore.command(ParserSchema.dC_PARSE, args,
    // parseFile);
    //
    //   // The parsing job is done on a service thread. We have to wait until
    //   // the parsing is finished.
    //   while (!status.getAttribute(DE.A_VALUE).equals("really done")) {
    //	Thread.currentThread().sleep(100);
    //   }
    //   
    //   // Retrieve the parsed source object under which the parsed information
    // is contained.
    //   DataElement files = _dataStore.find(theProject, DE.A_NAME,
    // ParserSchema.ParsedFiles, 2);
    //   String realFileName = null;
    //   try {
    //	realFileName = (new File(fileName)).getCanonicalPath();
    //   }
    //   catch (IOException e) {
    //	CResource.writeLog(CImporter.instance(), "load", e);
    //   }
    //
    //   DataElement parsedSource = _dataStore.find(files, DE.A_VALUE,
    // realFileName, 1);
    //
    //   // Terminate the parser background threads
    //   DataElement terminateParseThread = _dataStore.createObject(preferences,
    // "preferences", "Terminate Thread");
    //   DataElement cancelCommand =
    // _dataStore.find(_dataStore.getDescriptorRoot(), DE.A_VALUE, "C_CANCEL",
    // 2);
    //   _dataStore.command(cancelCommand, terminateParseThread, parseFile, true);
    //   
    //   // Throw an exception if the parsed source is null.
    //   if (parsedSource == null) {
    //	 throw new CException(CResource.getString("_ERROR_Parsed_Source_Is_Null",
    // realFileName));
    //   }
    //   
    //   // Depending on whether a type name is specified, call the corresponding
    // load() method
    //   // to translate the parsed information to MOF format.
    //   Resource resource = null;
    //   if (typeName == null)
    //	 resource = load(parsedSource);
    //   else
    //	 resource = load(parsedSource, typeName);
    //   
    //   // Disconnect from the datastore
    //   client.disconnect();
    //   
    //   return resource;
    // }

    /**
     * Initialize the values that are unique for each importer session. These
     * include the preference settings.
     */
    private void initialize(Map options) throws Exception {

        CPreferenceStore.getValuesFromPlugin();

        // Parse the options if it is not null.
        //   if (options != null) {
        //	// Parse the options as a String.
        //	if (options instanceof String)
        //	 CPreferenceStore.parseOptionsAsString((String)options);
        //	// Parse the options as key-value pairs
        //	else if (options instanceof HashMap)
        //	 CPreferenceStore.parseOptionsAsKey((HashMap)options);
        //   }

        if (options != null) {
            CPreferenceStore.parseOptionsAsKey((HashMap) options);
        }

        // Create the platform compiler info object
        CTypeDescriptorBuilder.createPlatformCompilerInfo();
        // The Type Descriptor map is unique for each importer session.
        CPrimitiveTypesMapper.createTypeDescriptorMap();
        // Set the bitfield layout rule
        CBitFieldHelper.setRule();

        // Parse the include paths
        _includePaths.clear();
        parseIncludePath(CPreferenceStore.getIncludePaths(), _includePaths);

        // By default, we use a new CModelConverter
        // object for every importer session. If the "-sameproject" option is
        // specified,
        // we will reuse the existing CModelConverter object and the existing
        // namespace.
        //
        if (!CPreferenceStore.isReuseParseInformation())
            //_currentModelConverter = new CModelConverter(this);
            _currentModelConverter = CModelConverter.createInstance();

    }

    /*
     * (non-Javadoc)
     * 
     * @see org.eclipse.emf.ecore.resource.impl.ResourceImpl#doLoad(java.io.InputStream,
     *      java.util.Map)
     */
    public void doLoad(InputStream arg0, Map arg1) throws IOException {
        try {
            initialize(arg1);

            //		_uri = uri;
            //
            //		// Throws an exception if the uri is null.
            //		if (uri == null || uri.length() == 0) {
            //		 throw new CException(CResource.getString("_ERROR_URI_Is_Null"));
            //		}
            //
            //		// Get the URIFactory
            //		URIFactory uriFactory = URIFactoryRegister.getFactory();
            //		if (uriFactory == null) {
            //		 throw new
            // CException(CResource.getString("_ERROR_URIFactory_Is_Null"));
            //		}

            String fileName = null;
            String typeName = null;
            //		try {
            //		 URI myURI = getURI();
            //		 java.net.URL myURL = myURI.getURL();
            //		 // If the URL protocol is "platform", we need to resolve it to
            //		 // the physical path.
            //		 if (myURL.getProtocol().equals("platform")) {
            //		  myURL = org.eclipse.core.runtime.Platform.resolve(myURL);
            //		 }
            //		 fileName = myURL.getFile();
            //		 typeName = myURL.getRef();
            //		}
            //		catch (Exception e) {
            //		 // If uri is not in a valid format of URI, treat it as a
            //		 // regular file name.
            //		 fileName = new String(_uri);
            //		 int index = _uri.indexOf('#');
            //		 if (index > 0) {
            //		  fileName = _uri.substring(0, index);
            //		  typeName = _uri.substring(index+1);
            //		 }
            //		}

            fileName = getURI().toFileString();
            int index = fileName.indexOf('#');
            if (index > 0) {
                String temp = fileName;
                fileName = temp.substring(0, index);
                typeName = temp.substring(index + 1);
            }

            // Get the absolute path name
            File file = new File(fileName);
            if (file.exists()) {
                fileName = file.getAbsolutePath();
            } else {
                throw new CException(CResource.getString("_ERROR_File_Not_Exist", fileName));
            }

            // TODO: integration of CDT V2 Parser
            boolean isCDT2Parser = true;

            if (isCDT2Parser) {
                CParseCallback cParseCallback = parseCSource(file);

                if (cParseCallback != null) {
                    // Depending on whether a type name is specified, call the
                    // corresponding load() method
                    // to translate the parsed information to EMF format.
                    if (typeName == null) {
                        doLoad(cParseCallback);
                    } else {
                        doLoad(cParseCallback, typeName);
                    }
                }

            } else {
                // use C Parser of CDT V1

                // Create a client connection which will create a DataStore
                // implicitly.
                ClientConnection client = new ClientConnection("Import");
                _dataStore = client.getDataStore();

                // Set the plugin path which is used to locate the data files
                // (e.g.
                // minerFile.dat)
                // needed by the cpp model.
                // This file is located at
                // %eclipsedir%/plugins/com.ibm.etools.c/.
                //

                String pluginPath = CPlugin.getInstallationDirectory(); //org.eclipse.core.boot.BootLoader.getInstallURL().getFile();
                //		  if (baseDir != null) {
                //			pluginPath = baseDir + "plugins/";
                //		  }
                //		  else {
                //			String eclipseDir = System.getProperty("eclipsedir");
                //			if (eclipseDir == null) {
                //			  eclipseDir = "c:/eclipse";
                //			}
                //		  pluginPath = eclipseDir + "/plugins/";
                //		  }

                _dataStore.setAttribute(DataStoreAttributes.A_PLUGIN_PATH, pluginPath);
                // the pluginpath now has everything it needs - set the location
                // to
                // blank
                _dataStore.setMinersLocation(""); //"com.ibm.etools.c");
                client.localConnect();

                // Initialize the schema and miners
                _dataStore.showTicket(null);
                _dataStore.getSchema();
                _dataStore.initMiners();

                // Create a DataElement to represent the object file we are
                // going to
                // parse.
                DataElement parseFile = _dataStore.createObject(_dataStore.getHostRoot(), "file", fileName, fileName);

                // Find the miner root under which the parsed information is
                // located.
                DataElement minerData = _dataStore.findMinerInformation("com.ibm.cpp.miners.parser.ParseMiner");

                // Create a fake project to hold the file
                DataElement theProject = _dataStore.createObject(minerData, ParserSchema.Project, "fake project");
                DataElement parsedFiles = _dataStore.createObject(theProject, ParserSchema.ParsedFiles, ParserSchema.ParsedFiles);
                _dataStore.createObject(theProject, ParserSchema.SystemObjects, ParserSchema.SystemObjects);
                _dataStore.createObject(theProject, ParserSchema.SystemObjects, ParserSchema.ProjectObjects);

                DataElement projectMinerProject = _dataStore.createObject(minerData, ParserSchema.Project, "fake project2");
                _dataStore.createReference(projectMinerProject, theProject, ParserSchema.ParseReference);

                // Set the include path
                DataElement preferences = _dataStore.createObject(theProject, "preferences", "Preferences");
                DataElement includeElement = _dataStore.createObject(preferences, "preferences", "Include Path");
                for (int i = 0; i < _includePaths.size(); i++)
                    _dataStore.createObject(includeElement, "directory", (String) _includePaths.get(i), (String) _includePaths.get(i));

                // Invoke the C/C++ parser on the project
                ArrayList args = new ArrayList();
                args.add(projectMinerProject);
                DataElement status = _dataStore.command(ParserSchema.dC_PARSE, args, parseFile);

                // The parsing job is done on a service thread. We have to wait
                // until
                // the parsing is finished.
                while (!status.getAttribute(DE.A_VALUE).equals("really done")) {
                    //Thread.currentThread().sleep(100);
                    Thread.sleep(100);
                }

                // Retrieve the parsed source object under which the parsed
                // information is contained.
                DataElement files = _dataStore.find(theProject, DE.A_NAME, ParserSchema.ParsedFiles, 2);
                String realFileName = null;
                try {
                    realFileName = (new File(fileName)).getCanonicalPath();
                } catch (IOException e) {
                    CResource.writeLog(this, "doLoad", e);
                }

                DataElement parsedSource = _dataStore.find(files, DE.A_VALUE, realFileName, 1);

                // Terminate the parser background threads
                DataElement terminateParseThread = _dataStore.createObject(preferences, "preferences", "Terminate Thread");
                DataElement cancelCommand = _dataStore.find(_dataStore.getDescriptorRoot(), DE.A_VALUE, "C_CANCEL", 2);
                _dataStore.command(cancelCommand, terminateParseThread, parseFile, true);

                // Throw an exception if the parsed source is null.
                if (parsedSource == null) {
                    throw new CException(CResource.getString("_ERROR_Parsed_Source_Is_Null", realFileName));
                }

                // Depending on whether a type name is specified, call the
                // corresponding load() method
                // to translate the parsed information to MOF format.
                if (typeName == null)
                    doLoad(parsedSource);
                else
                    doLoad(parsedSource, typeName);

                // Disconnect from the datastore
                client.disconnect();
            }
        } catch (Exception e) {
            throw new IOException(e.getLocalizedMessage());
        }
    }

    /***************************************************************************
     * 
     * sets the ID into the passed Object. ID is the name of the Object,
     * 
     * @param Object
     *            for which the ID must be set
     *  
     **************************************************************************/
    public void setRefIDs(Object obj) {

        String typeName = null;
        String elementName = null;
        String name = null;

        if (obj instanceof CStructured || obj instanceof CTypedef) {

            if (obj instanceof CStructured) {
                name = ((CStructured) obj).getName();
            } else if (obj instanceof CTypedef) {
                name = ((CTypedef) obj).getName();
            }

            if (name != null && !name.equals("")) {
                typeName = "Type:" + name;
                elementName = "Element:" + name;

                setID((EObject) obj, name);
            }
        }
    }

    /**
     * TODO: integration of CDT V2 Parser
     * 
     * Parse passed C Source file with C Parser of CDT V2
     * 
     * @param File
     *            C source file to be parsed
     * 
     * @return CParseCallback contains parsed C source file
     * 
     * @exception CException
     *                will be thrown if any error occured
     *  
     */
    public CParseCallback parseCSource(File file) throws CException {

        CParseCallback quickParseCallback = null;
        IASTCompilationUnit compUnit = null;

        String code = ""; //$NON-NLS-1$

        try {
            String filePath = file.getCanonicalPath();

            //
            // use structural parse mode
            //
            ParserMode mode = ParserMode.STRUCTURAL_PARSE;
            //ParserMode mode = ParserMode.COMPLETE_PARSE;

            //
            // create parse callback
            //
            quickParseCallback = new CParseCallback();
            quickParseCallback.setFileName(file.getName());

            //
            // pick the language
            //
            ParserLanguage language = ParserLanguage.C;

            //
            // create the parser
            //
            IParser parser = null;

            try {

                //
                // get include path for C parser
                //
                String[] includePathArray = new String[_includePaths.size()];
                Iterator includePathIterator = _includePaths.iterator();
                int arrayIndex = 0;

                while (includePathIterator.hasNext()) {
                    includePathArray[arrayIndex] = (String) (includePathIterator.next());
                    arrayIndex++;
                }

                //
                // create parser
                //
                IScannerInfo scanInfo = new ScannerInfo(new HashMap(), includePathArray);

                CodeReader codeReader = new CodeReader(filePath);

                IScanner scanner = ParserFactory.createScanner(codeReader, scanInfo, mode, language, quickParseCallback, null, null);

                parser = ParserFactory.createParser(scanner, quickParseCallback, mode, language, null);
            } catch (ParserFactoryError ex) {
                throw new CException(Messages.getString("_ERROR_Parsing_File", new String[] { filePath, ex.getMessage() }));
            }

            //
            // now parse C Source file
            //
            boolean hasNoErrors = parser.parse();

            if (!hasNoErrors) {

                //
                // in case of an error get all problems from callback and trow
                // exception
                //
                String problemMessage = "";
                Iterator iterator = quickParseCallback.getProblems();

                while (iterator.hasNext()) {
                    IProblem problem = (IProblem) iterator.next();
                    problemMessage += problem.getMessage();
                    problemMessage += "\n";
                }

                throw new CException(Messages.getString("_ERROR_C_Parser", new Object[] { filePath, problemMessage }));
            }

        } catch (IOException ex) {
            throw new CException(Messages.getString("_ERROR_Parsing_File", new String[] { file.getName(), ex.getMessage() }));
        }

        return quickParseCallback;
    }

    /**
     * 
     * Translate the parsed output saved in the CParseCallback tree into EMF
     * resource format. The return result is a EMF resource object whose Extent
     * is populated with the parsed object list.
     * 
     * @param CParseCallback
     *            parsed_source
     * 
     * @exception Exception
     *                if an error occurred
     */
    protected void doLoad(CParseCallback cParseCallback) throws Exception {

        ArrayList objectList = _currentModelConverter.importSource(cParseCallback);

        ArrayList finalObjectList = objectList;

        if (CPreferenceStore.isGenerateFullObjectList()) {
            finalObjectList = CUtility.generateFullObjectList(objectList);
        }

        getContents().addAll(finalObjectList);

        if (finalObjectList.size() <= 0) {
            throw new CException(Messages.getString("_WARNING_No_Structure", cParseCallback.getFileName()));
        }
    }

    /**
     * 
     * TODO: integration of CDT V2 Parser
     * 
     * Retrieve the named type from the parsed file Only the objects related to
     * that type are added to the returned list
     * 
     * @param CParseCallback
     *            parsed_source
     * @param String
     *            type name to be retrieved
     * 
     * @exception Exception
     *                if an error occurred
     *  
     */
    protected void doLoad(CParseCallback cParseCallback, String typeName) throws Exception {

        if (_currentModelConverter.getSourceUnitFromName(cParseCallback.getFileName()) == null) {
            _currentModelConverter.importSource(cParseCallback);
        }

        ArrayList objectList = new ArrayList();
        CTypeTable typeTable = _currentModelConverter.getTypeTable();
        Object modelType = typeTable.getType(typeName);

        if (modelType != null) {
            objectList.add(modelType);
        } else {
            //
            // If we cannot find a type for the given name, try to find it as a
            // function
            //
            Object modelFunction = typeTable.getFunction(typeName);
            if (modelFunction != null) {
                objectList.add(modelFunction);
            } else
                throw new CException(CResource.getString("_ERROR_No_Type_Or_Function", typeName));
        }

        ArrayList finalObjectList = objectList;

        if (CPreferenceStore.isGenerateFullObjectList()) {
            finalObjectList = CUtility.generateFullObjectList(objectList);
        }

        getContents().addAll(finalObjectList);
    }
}