/*
 * Licensed Materials - Property of IBM
 *
 * com.ibm.etools.c
 *
 *  Copyright IBM Corporation 2004. All Rights Reserved. 
 *
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.etools.c.importer;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

import com.ibm.etools.c.CArray;
import com.ibm.etools.c.CClassifier;
import com.ibm.etools.c.CDatatype;
import com.ibm.etools.c.CDerivableType;
import com.ibm.etools.c.CField;
import com.ibm.etools.c.CPointer;
import com.ibm.etools.c.CStruct;
import com.ibm.etools.c.CStructureContents;
import com.ibm.etools.c.CStructured;
import com.ibm.etools.c.CTypedElement;
import com.ibm.etools.c.CTypedef;
import com.ibm.etools.c.CUnion;
import com.ibm.etools.c.datatypes.CBitField;
import com.ibm.etools.c.datatypes.CChar;
import com.ibm.etools.c.datatypes.CDouble;
import com.ibm.etools.c.datatypes.CEnumeration;
import com.ibm.etools.c.datatypes.CFloat;
import com.ibm.etools.c.datatypes.CInt;
import com.ibm.etools.c.datatypes.CLong;
import com.ibm.etools.c.datatypes.CLongDouble;
import com.ibm.etools.c.datatypes.CLongLong;
import com.ibm.etools.c.datatypes.CShort;
import com.ibm.etools.c.datatypes.CUnsignedLong;
import com.ibm.etools.c.datatypes.CUnsignedLongLong;
import com.ibm.etools.c.datatypes.CUnsignedShort;
import com.ibm.etools.c.datatypes.CWchar;

/**
 * CTypeDescriptorHelper is a helper class for the population of the Type
 * Descriptor model. It contains helper methods to calculate the size and
 * alignment requirement of a C model type and the offset of a struct field.
 */
public class CTypeDescriptorHelper {

    public static final String copyright = "Licensed Material - Property of IBM\n com.ibm.etools.c\n(C) Copyright IBM Corp. 2004 - All Rights Reserved.\nUS Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp."; //$NON-NLS-1$

    private static CTypeDescriptorHelper _instance = new CTypeDescriptorHelper();

    /**
     * Constructor
     */
    private CTypeDescriptorHelper() {
    }

    /**
     * Return the only instance of CTypeDescriptorHelper
     */
    public static CTypeDescriptorHelper instance() {
        return _instance;
    }

    /**
     * Return the size of a C type
     * 
     * @return The size of the given type
     * @param A
     *            C model type
     */
    public static int sizeof(CClassifier type) {

        if (type == null) {
            CResource.writeLog(CTypeDescriptorHelper.instance(), "sizeof", CResource.getString("_ERROR_Type_Is_Null_For_Sizeof"));
            return 0;
        }

        try {
            if (type instanceof CPointer)
                return sizeofPointer((CPointer) type);
            else if (type instanceof CArray)
                return sizeofArray((CArray) type);
            else if (type instanceof CStruct)
                return sizeofStruct((CStruct) type);
            else if (type instanceof CUnion)
                return sizeofUnion((CUnion) type);
            else if (type instanceof CTypedef)
                return sizeofTypedef((CTypedef) type);
            else if (type instanceof CDatatype)
                return sizeofPrimitiveType((CDatatype) type);
            else
                return 0;
        } catch (Exception e) {
            CResource.writeLog(CTypeDescriptorHelper.instance(), "sizeof", e);
            return 0;
        }
    }

    /*
     * Return the size of a typed element (field or parameter). This will call
     * the sizeOf() method on the type of the typed element.
     */
    public static int sizeof(CTypedElement element) {
        return sizeof(CQuery.typeOf(element));
    }

    /**
     * Return the size of a C primitive type
     */
    private static int sizeofPrimitiveType(CDatatype type) throws Exception {

        if (type instanceof CShort || type instanceof CUnsignedShort)
            return 2;
        else if (type instanceof CLongLong || type instanceof CUnsignedLongLong || type instanceof CDouble)
            return 8;
        else if (type instanceof CLong || type instanceof CUnsignedLong) {
            if (CPreferenceStore.is64bitAddress())
                return 8;
            else
                return 4;
        } else if (type instanceof CInt || type instanceof CFloat) //@P1C
            return 4;
        else if (type instanceof CEnumeration) //@P1A
            return CTypeDescriptorBuilder.calculateEnumSize((CEnumeration) type); //@P1A
        else if (type instanceof CWchar) //@P2A
            return CPreferenceStore.getSizeWchar(); //@P2A
        else if (type instanceof CChar)
            return 1;
        else if (type instanceof CLongDouble) {
            if (CPreferenceStore.isLongDouble128())
                return 16;
            else
                return 8;
        }
        // Be careful that the bitfield size is in bits.
        else if (type instanceof CBitField)
            return sizeof(((CBitField) type).getBaseType());
        else
            return 0;
    }

    /**
     * Return the size of a pointer type. A pointer is 8 bytes on 64bit platform
     * and 4 bytes on 32bit platform.
     */
    private static int sizeofPointer(CPointer apointer) {
        if (CPreferenceStore.is64bitAddress())
            return 8;
        else
            return 4;
    }

    /**
     * Return the size of a typedef The size of a typedef equals to the size of
     * its reference type.
     */
    private static int sizeofTypedef(CTypedef atypedef) throws Exception {
        CDerivableType referenceType = atypedef.getDerives();
        if (referenceType instanceof CClassifier)
            return sizeof((CClassifier) referenceType);
        else
            throw new CException(CResource.getString("_ERROR_Typedef_Basetype_Is_Not_valid", atypedef.toString()));
    }

    /**
     * Return the size of an array. The size of an array equals to the size of
     * one element multiplied by the number of elements.
     */
    private static int sizeofArray(CArray array) throws Exception {

        int dimension = array.getDimension().intValue();
        if (dimension < 1) {
            throw new CException(CResource.getString("_ERROR_Array_Size_Not_Found", array.toString()));
        }
        CDerivableType referenceType = array.getDerives();
        if (referenceType instanceof CClassifier) {
            return sizeof((CClassifier) referenceType) * dimension;
        } else {
            throw new CException(CResource.getString("_ERROR_Array_Basetype_Not_Valid", array.toString()));
        }
    }

    //Defect24439: The following Hashtable was added to stop recursion while
    //processing recursive structs (structs containing fields of type matching
    // the parent struct)
    private static Hashtable _names = new Hashtable();

    /**
     * Find the details for the structure, including the size and alignment of
     * the struct, the size and offset of each field.
     */
    private static CStructuredDetails findStructDetails(CStruct astruct) throws CException {

        //Defect24439: Make sure we aren't asking to find the StructDetails of
        //struct currently being evaluated.
        String structName = astruct.getName();
        if (_names.containsKey(structName))
            return (CStructuredDetails) _names.get(structName);

        CStructuredDetails details = new CStructuredDetails(astruct);

        //Defect24439: Store the name and the details of the Struct
        _names.put(structName, details);

        List fields = astruct.getContains();
        int numberOfFields = fields.size();
        int size = 0;
        int alignment = 0;

        //CBitFieldHelper bitfieldHelper = null;
        for (int i = 0; i < numberOfFields; i++) {
            Object afield = fields.get(i);
            int fieldSize = 0;
            int fieldAlignment = 0;
            boolean isBitField = false;

            CClassifier fieldType = null;
            if (afield instanceof CField) {
                fieldType = CQuery.typeOf((CField) afield);
                if (fieldType instanceof CBitField) {
                    isBitField = true;
                    CBitFieldHelper.parse(astruct, (CField) afield, i, size);
                    fieldSize = CBitFieldHelper.getSize();
                    fieldAlignment = CBitFieldHelper.getAlignment();
                } else {
                    fieldSize = sizeof((CField) afield);
                    fieldAlignment = alignmentOf((CField) afield);
                }
            } else if (afield instanceof CStructured) {
                fieldSize = sizeof((CStructured) afield);
                fieldAlignment = alignmentOf((CStructured) afield);
            }

            if (fieldAlignment > alignment)
                alignment = fieldAlignment;

            if (isBitField) {
                size = CBitFieldHelper.getCurrentStructSize();
                int sizeInBits = ((CBitField) fieldType).getSize().intValue();
                int offsetInBits = CBitFieldHelper.getOffset();
                details.setFieldSize(i, sizeInBits);
                details.setFieldOffset(i, offsetInBits);
            } else {
                details.setFieldSize(i, fieldSize);

                // Calculate the padding and the new size.
                if (fieldAlignment == 0 || size % fieldAlignment == 0) {
                    details.setFieldOffset(i, size);
                    size = size + fieldSize;
                } else {
                    details.setFieldOffset(i, size + fieldAlignment - size % fieldAlignment);
                    size = size + fieldSize + (fieldAlignment - size % fieldAlignment);
                }
            }
        }

        // The alignment cannot be greater than the pack level.
        int packLevel = CPreferenceStore.getPackLevel();
        if (alignment > packLevel)
            alignment = packLevel;

        details.setAlignment(alignment);

        // Adjust the struct size to be multiples of the alignment.
        if (!(alignment == 0 || size % alignment == 0))
            size = size + alignment - size % alignment;

        details.setSize(size);

        // Add the structure details to the type table.
        addStructuredDetails(astruct, details);

        //Defect24439: If we get here, we are done processing the struct, so we
        // can remove it
        _names.remove(structName);

        return details;
    }

    /**
     * Find the details for the union, including the size and alignment of the
     * union, the size and offset of each field.
     */
    private static CStructuredDetails findUnionDetails(CUnion aunion) throws CException {

        CStructuredDetails details = new CStructuredDetails(aunion);

        List fields = aunion.getContains();
        int numberOfFields = fields.size();
        int size = 0;
        int alignment = 0;

        for (int i = 0; i < numberOfFields; i++) {
            Object afield = fields.get(i);
            int fieldSize = 0;
            int fieldAlignment = 0;

            if (afield instanceof CField) {
                fieldSize = sizeof((CField) afield);
                fieldAlignment = alignmentOf((CField) afield);
            } else if (afield instanceof CStructured) {
                fieldSize = sizeof((CStructured) afield);
                fieldAlignment = alignmentOf((CStructured) afield);
            }

            details.setFieldSize(i, fieldSize);
            details.setFieldOffset(i, 0);

            // The size of the union is the maximum of the size of all the
            // fields.
            if (fieldSize > size)
                size = fieldSize;

            if (fieldAlignment > alignment)
                alignment = fieldAlignment;
        }

        // The alignment cannot be greater than the pack level.
        int packLevel = CPreferenceStore.getPackLevel();
        if (alignment > packLevel)
            alignment = packLevel;

        details.setAlignment(alignment);
        details.setSize(size);

        // Add the union details to the type table.
        addStructuredDetails(aunion, details);

        return details;
    }

    /**
     * Return the size of the structure The size is retrieved from the
     * CStructuredDetails object.
     */
    private static int sizeofStruct(CStruct astruct) throws CException {
        CStructuredDetails details = getStructuredDetails(astruct);
        if (details == null)
            details = findStructDetails(astruct);

        return details.getSize();
    }

    /**
     * Return the size of the union The size is retrieved from the
     * CStructuredDetails object.
     */
    private static int sizeofUnion(CUnion aunion) throws CException {
        CStructuredDetails details = getStructuredDetails(aunion);
        if (details == null)
            details = findUnionDetails(aunion);

        return details.getSize();
    }

    /**
     * Return the offset of a CStructureContents within its container (a struct
     * or union). The structure content object can be either a field or a nested
     * struct/union.
     * 
     * @return The offset of the given structure content
     * @param A
     *            structure content object
     */
    public static int offsetOf(CStructureContents structureContents) throws CException {

        int offset = 0;
        CStructured container = structureContents.getContainer();
        if (container == null)
            return 0;

        int baseOffset = 0;
        if (container.getContainer() != null)
            baseOffset = offsetOf(container);

        List fields = container.getContains();
        int i;
        for (i = 0; i < fields.size(); i++) {
            if (fields.get(i) == structureContents)
                break;
        }

        CStructuredDetails details = getStructuredDetails(container);

        if (details != null && i < fields.size()) {
            offset = details.getFieldOffset(i);
        }

        offset = baseOffset + offset;
        return offset;
    }

    /**
     * Return the alignment requirement for a C type. Rule to calculate the
     * alignment: 1) The alignment of a typedef equals to the alignment of its
     * referenced type; 2) The alignment of an array equals to the alignment of
     * one element; 3) The alignment of a struct/union equals to the maximum
     * alignment of its fields; 4) In other cases, the alignment equals to the
     * size of the type.
     * 
     * If the pack level is specified, the alignment is the minimum of the above
     * value and the pack level.
     */
    public static int alignmentOf(CClassifier type) throws CException {

        if (type == null)
            return 0;

        // If the pack level is 1, the alignment of any type is always 1.
        int packLevel = CPreferenceStore.getPackLevel();
        if (packLevel == 1)
            return 1;

        int alignment = 0;

        CClassifier baseType = CQuery.baseType(type);
        if (baseType == null) {
            return 0;
        }

        if (baseType instanceof CStructured)
            alignment = alignmentOfStructured((CStructured) baseType);
        else
            alignment = sizeof(baseType);

        // Alignment cannot be greater than the pack level.
        if (alignment > packLevel)
            alignment = packLevel;

        return alignment;
    }

    /**
     * Return the alignment of a struct/union field Normally the alignment of a
     * field equals to the alignment of its type.
     * 
     * Power alignment needs special treatment: A field whose type is double or
     * long double is aligned at word boundary if it is not the first field of
     * the struct/union.
     */
    public static int alignmentOf(CField afield) throws CException {
        CClassifier fieldType = CQuery.typeOf(afield);
        if (CPreferenceStore.getAlignment() == CPreferenceStore.POWER_ALIGNMENT && !CQuery.isFirstField(afield) && (fieldType instanceof CDouble || fieldType instanceof CLongDouble)) {
            return 4;
        } else {
            return alignmentOf(fieldType);
        }
    }

    /**
     * Return the alignment requirement of a C structure or union. The alignment
     * requirement of a struct/union is the maximum of the alignment
     * requirements of its fields. A struct/union member can be a regular field,
     * a nested struct/union or an enumeration.
     */
    private static int alignmentOfStructured(CStructured astructured) throws CException {

        int packLevel = CPreferenceStore.getPackLevel();
        List fields = astructured.getContains();
        Iterator it = fields.iterator();
        int maxAlignment = 0;
        int fieldAlignment;
        while (it.hasNext()) {
            Object afield = it.next();

            //Defect24439: If the name currently being processed is in the
            // _names list, then abort.
            String theStructName = astructured.getName();
            if ((theStructName != null) && (!theStructName.equals("")) && (_names.containsKey(astructured.getName()))) {
                _names.clear();
                throw new CException(CResource.getString("_ERROR_Unknown_Size_Type", astructured.getName()));
            }
            if ((afield instanceof CField) && !(_names.containsKey(astructured.getName())))
                fieldAlignment = alignmentOf((CField) afield);
            else if (afield instanceof CStructured)
                fieldAlignment = alignmentOf((CStructured) afield);
            else
                fieldAlignment = 0;

            if (fieldAlignment >= packLevel)
                return packLevel;
            else if (fieldAlignment > maxAlignment)
                maxAlignment = fieldAlignment;
        }

        return maxAlignment;
    }

    /**
     * Return the CStructuredDetails object for a struct/union
     */
    public static CStructuredDetails getStructuredDetails(CStructured astructured) throws CException {
        //CTypeTable typeTable =
        // CImporter.instance().getCurrentModelConverter().getTypeTable();
        CTypeTable typeTable = CModelConverter.getInstance().getTypeTable();
        CStructuredDetails details = typeTable.getStructuredDetails(astructured);
        if (details == null) {
            if (astructured instanceof CStruct)
                details = findStructDetails((CStruct) astructured);
            else if (astructured instanceof CUnion)
                details = findUnionDetails((CUnion) astructured);
        }

        return details;
    }

    /**
     * Set the CStructuredDetails object for a struct/union
     */
    public static void addStructuredDetails(CStructured astructured, CStructuredDetails details) {
        //CTypeTable typeTable =
        // CImporter.instance().getCurrentModelConverter().getTypeTable();
        CTypeTable typeTable = CModelConverter.getInstance().getTypeTable();
        typeTable.addStructuredDetails(astructured, details);
    }

}